/* Copyright (C) logi.cals GmbH. All rights reserved. */

#include <lcfu_iec61131__EXPT.h>
#include <inttypes.h>
#include <stdint.h>
#include <limits.h>

/*
 ** recursive function to calculate base**exp, method is divide and conquer
 ** returns 0 on overflow (over ulonglong max)
 ** does not handle singed values!
 */
static unsigned long long integerPow(unsigned long long base, unsigned exp)
{
  unsigned long long a;
  if (exp == 1)
  {
    return base;
  }
  a = integerPow(base, exp / 2);
  if (a == 0)
  {
    return 0;
  }
  if (a >= ULLONG_MAX / a)
  {
    return 0;
  }
  a *= a;

  if (exp % 2 == 0)
  {
    return a;
  }
  if (a >= ULLONG_MAX / base)
  {
    return 0;
  }
  return a * base;
}

/*
 ** impl for a**b for signed values, range up to maxval
 */
signed long long LC_MF_EXPT_SINTEGER_IMPL(unsigned long long maxval, signed long long in1, signed long long in2, LC_TD_BOOL *pENO)
{
  unsigned long long underflowMaxVal = maxval / 2 + 1;
  unsigned long long overflowMaxVal = maxval / 2;
  unsigned char isNeg = 0;
  LC_TD_BOOL localEno;
  unsigned long long result;

  if (pENO == NULL)
  {
    pENO = &localEno;
  }

  if ((in2 == 0) || (in1 == 1))
  {
    return 1;
  }
  if (in1 == -1)
  {
    if (in2 % 2 == 0)
    {
      return 1;
    }
    return -1;
  }
  if (in1 == 0)
  {
    if (in2 > 0)
    {
      return 0;
    }
    *pENO = LC_EL_false;
    return overflowMaxVal;
  }
  if (in2 < 0)
  {
    return 0;
  }
  if (in2 == 1)
  {
    return in1;
  }

  if (in1 == underflowMaxVal)
  {
    *pENO = LC_EL_false;
    return in1;
  }

  if (in1 < 0)
  {
    if (in2 % 2 != 0)
    {
      isNeg = 1;
    }
    in1 = -in1;
  }

  /* 2**64 is not more possible. so max exponent is 63 */
  if (in2 >= 63)
  {
    *pENO = LC_EL_false;
    if (!isNeg)
    {
      return overflowMaxVal;
    }
    /* special handling for long-long-min necessary. - else compilers change it to 0 */
    if (maxval == ULLONG_MAX)
    {
      return LLONG_MIN;
    }
    return (unsigned long long) (-(long) underflowMaxVal);
  }

  result = integerPow((unsigned long long) in1, (unsigned) in2);
  if (isNeg && (result > underflowMaxVal || result == 0))
  {
    *pENO = LC_EL_false;
    /* special handling for long-long-min necessary. - else compilers change it to 0 */
    if (maxval == ULLONG_MAX)
    {
      return LLONG_MIN;
    }
    return (unsigned long long) (-(long) underflowMaxVal);
  }
  if ((!isNeg) && (result > overflowMaxVal || result == 0))
  {
    *pENO = LC_EL_false;
    return overflowMaxVal;
  }

  if (isNeg)
  {
    return (unsigned long long) (-(long) result);
  }
  return result;
}

/*
 ** impl for a**b for unsigned values, range up to maxval
 */
unsigned long long LC_MF_EXPT_UINTEGER_IMPL(unsigned long long maxval, unsigned long long in1, unsigned long long in2, LC_TD_BOOL *pENO)
{
  LC_TD_BOOL localEno;
  unsigned long long result;

  if (pENO == NULL)
  {
    pENO = &localEno;
  }

  if ((in2 == 0) || (in1 == 1))
  {
    return 1;
  }
  if (in1 == 0)
  {
    return 0;
  }
  if (in2 == 1)
  {
    return in1;
  }

  /* 2**64 is not more possible. so max exponent is 63 */
  if (in2 >= 63)
  {
    *pENO = LC_EL_false;
    return maxval;
  }
  result = integerPow(in1, (unsigned) in2);
  if (result == 0)
  {
    *pENO = LC_EL_false;
    return maxval;
  }

  if (result > maxval || result == 0)
  {
    *pENO = LC_EL_false;
    return maxval;
  }

  return result;
}
